@@ -28,6 +28,8 @@ var isInstallable = require('./validate-args.js').isInstallable
 var packageId = require('../utils/package-id.js')
 var moduleName = require('../utils/module-name.js')
 
+exports.test = {} // used to hold functions for testing by unit tests
+
 // The export functions in this module mutate a dependency tree, adding
 // items to them.
 
@@ -125,12 +127,22 @@ function recalculateMetadata (tree, log, seen, next) {
 function addRequiredDep (tree, child) {
   if (!isDep(tree, child)) return false
   var name = isProdDep(tree, moduleName(child)) ? flatNameFromTree(tree) : '#DEV:' + flatNameFromTree(tree)
-  child.package._requiredBy = union(child.package._requiredBy || [], [name])
-  child.requiredBy = union(child.requiredBy || [], [tree])
-  tree.requires = union(tree.requires || [], [child])
+  replaceModuleName(child.package, '_requiredBy', name)
+  replaceModule(child, 'requiredBy', tree)
+  replaceModule(tree, 'requires', child)
   return true
 }
 
+function removeObsoleteDep (child) {
+  if (child.removed) return
+  child.removed = true
+  var requires = child.requires || []
+  requires.forEach(function (requirement) {
+    requirement.requiredBy = requirement.requiredBy.filter(function (reqBy) { reqBy !== child })
+    if (requirement.requiredBy.length === 0) removeObsoleteDep(requirement)
+  })
+}
+
 function matchingDep (tree, name) {
   if (tree.package.dependencies && tree.package.dependencies[name]) return tree.package.dependencies[name]
   if (tree.package.devDependencies && tree.package.devDependencies[name]) return tree.package.devDependencies[name]
@@ -176,10 +188,6 @@ exports.loadRequestedDeps = function (args, tree, saveToDependencies, log, next)
   validate('AOOF', [args, tree, log, next])
   asyncMap(args, function (pkg, done) {
     var depLoaded = andAddParentToErrors(tree, done)
-    var pkgName = moduleName(pkg)
-    tree.children = tree.children.filter(function (child) {
-      return moduleName(child) !== pkgName
-    })
     resolveWithNewModule(pkg, tree, log.newGroup('loadRequestedDeps'), iferr(depLoaded, function (child, tracker) {
       validate('OO', arguments)
       if (npm.config.get('global')) {
@@ -201,7 +209,7 @@ exports.loadRequestedDeps = function (args, tree, saveToDependencies, log, next)
       // won't be when we're done), flag it as "depending" on the user
       // themselves, so we don't remove it as a dep that no longer exists
       if (!addRequiredDep(tree, child)) {
-        child.package._requiredBy = union(child.package._requiredBy, ['#USER'])
+        replaceModuleName(child.package, '_requiredBy', '#USER')
         child.directlyRequested = true
       }
       depLoaded(null, child, tracker)
@@ -209,6 +217,14 @@ exports.loadRequestedDeps = function (args, tree, saveToDependencies, log, next)
   }, andForEachChild(loadDeps, andFinishTracker(log, next)))
 }
 
+function moduleNameMatches (name) {
+  return function (child) { return moduleName(child) === name }
+}
+
+function noModuleNameMatches (name) {
+  return function (child) { return moduleName(child) !== name }
+}
+
 // while this implementation does not require async calling, doing so
 // gives this a consistent interface with loadDeps et al
 exports.removeDeps = function (args, tree, saveToDependencies, log, next) {
@@ -216,13 +232,13 @@ exports.removeDeps = function (args, tree, saveToDependencies, log, next) {
   args.forEach(function (pkg) {
     var pkgName = moduleName(pkg)
     if (saveToDependencies) {
-      var toRemove = tree.children.filter(function (child) { return moduleName(child) === pkgName })
-      tree.removed = union(tree.removed || [], toRemove)
+      var toRemove = tree.children.filter(moduleNameMatches(pkgName))
+      replaceModule(tree, 'removed', toRemove[0])
       toRemove.forEach(function (parent) {
         parent.save = saveToDependencies
       })
     }
-    tree.children = tree.children.filter(function (child) { return moduleName(child) !== pkgName })
+    tree.children = tree.children.filter(noModuleNameMatches(pkgName))
   })
   log.finish()
   next()
@@ -264,7 +280,7 @@ var failedDependency = exports.failedDependency = function (tree, name_pkg) {
     name = moduleName(pkg)
   }
 
-  tree.children = tree.children.filter(function (child) { return moduleName(child) !== name })
+  tree.children = tree.children.filter(noModuleNameMatches(name))
 
   if (isDepOptional(tree, name)) {
     return false
@@ -290,7 +306,7 @@ function andHandleOptionalErrors (log, tree, name, done) {
     if (!er) return done(er, child, childLog)
     var isFatal = failedDependency(tree, name)
     if (er && !isFatal) {
-      tree.children = tree.children.filter(function (child) { return moduleName(child) !== name })
+      tree.children = tree.children.filter(noModuleNameMatches(name))
       log.warn('install', "Couldn't install optional dependency:", er.message)
       log.verbose('install', er.stack)
       return done()
@@ -400,6 +416,29 @@ function flatNameFromTree (tree) {
   return flatName(path, tree)
 }
 
+exports.test.replaceModuleName = replaceModuleName
+function replaceModuleName (obj, key, name) {
+  validate('OSS', arguments)
+  obj[key] = union(obj[key] || [], [name])
+}
+
+exports.test.replaceModule = replaceModule
+function replaceModule (obj, key, child) {
+  validate('OSO', arguments)
+  if (!obj[key]) obj[key] = []
+  // we replace children with a new array object instead of mutating it
+  // because mutating it results in weird failure states.
+  // I would very much like to know _why_ this is. =/
+  var children = [].concat(obj[key])
+  var childName = moduleName(child)
+  for (var replaceAt = 0; replaceAt < children.length; ++replaceAt) {
+    if (moduleName(children[replaceAt]) === childName) break
+  }
+  var replacing = children.splice(replaceAt, 1, child)
+  obj[key] = children
+  return replacing[0]
+}
+
 function resolveWithNewModule (pkg, tree, log, next) {
   validate('OOOF', arguments)
   if (pkg.type) {
@@ -416,8 +455,6 @@ function resolveWithNewModule (pkg, tree, log, next) {
     }))
   }
 
-  var pkgName = moduleName(pkg)
-
   if (!pkg._from) {
     pkg._from = pkg._requested.name + '@' + pkg._requested.spec
   }
@@ -433,8 +470,8 @@ function resolveWithNewModule (pkg, tree, log, next) {
         isLink: tree.isLink
       })
 
-      parent.children = parent.children.filter(function (pkg) { return pkgName !== moduleName(pkg) })
-      parent.children.push(child)
+      var replaced = replaceModule(parent, 'children', child)
+      if (replaced) removeObsoleteDep(replaced)
       addRequiredDep(tree, child)
       pkg._location = flatNameFromTree(child)
 
@@ -481,7 +518,7 @@ function validateAllPeerDeps (tree, onInvalid, seen) {
 var findRequirement = exports.findRequirement = function (tree, name, requested) {
   validate('OSO', arguments)
   var nameMatch = function (child) {
-    return moduleName(child) === name && child.parent
+    return moduleName(child) === name && child.parent && !child.removed
   }
   var versionMatch = function (child) {
     return doesChildVersionMatch(child, requested)
@@ -509,15 +546,16 @@ var findRequirement = exports.findRequirement = function (tree, name, requested)
 // If it is, then it's the level below where its installed.
 var earliestInstallable = exports.earliestInstallable = function (requiredBy, tree, pkg) {
   validate('OOO', arguments)
-  var nameMatch = function (child) {
-    return moduleName(child) === pkg.name
-  }
 
-  if (tree.children.some(nameMatch)) return null
+  function undeletedModuleMatches (child) {
+    return !child.removed && moduleName(child) === pkg.name
+  }
+  if (tree.children.some(undeletedModuleMatches)) return null
 
   // If any of the children of this tree have conflicting
   // binaries then we need to decline to install this package here.
   var binaryMatches = tree.children.some(function (child) {
+    if (child.removed) return false
     return Object.keys(child.package.bin || {}).some(function (bin) {
       return pkg.bin && pkg.bin[bin]
     })
@@ -527,7 +565,8 @@ var earliestInstallable = exports.earliestInstallable = function (requiredBy, tr
   // if this tree location requested the same module then we KNOW it
   // isn't compatible because if it were findRequirement would have
   // found that version.
-  if (requiredBy !== tree && tree.package.dependencies && tree.package.dependencies[pkg.name]) {
+  var deps = tree.package.dependencies || {}
+  if (!tree.removed && requiredBy !== tree && deps[pkg.name]) {
     return null
   }
 
